\WW.XPL		23-Oct-2003	Loren Blaney	loren_blaney@idcomm.com
\Lorenzian Waterwheel using flip screen animation
\Compile with 16-bit XPL0

def	Buckets = 8;	\number of buckets in waterwheel
def 	BWidth=50, BHeight=55, B2Width=BWidth/2;   \bucket dimensions (pixels)
def	Base = 70;	\Y coordinate for baseline (X axis) of speed graph
def	SLimit = Base-13; \limit for speed graph (so it doesn't mess up screen)
def	ScnA=140+1, ScnB=ScnA+339+13; \starting Y coords for alternating screens

real	Speed,	\speed of waterwheel rotation (radians per iteration)
	Fric,	\friction slowdown rate (percent slowdown per iteration)
	WaterRate, \water fill rate (arbitrary units)
	Leak,	\bucket leak rate (percent leaked per iteration)
	Ang0,	\angle of reference point (bucket 0) on wheel (radians)
	AngB,	\angle between buckets around wheel (radians)
	Torque,	\torque on wheel caused by weight of water in buckets

	Water(Buckets),	\amount of water in each bucket
	BucX(Buckets),	\X coordinate of center of bucket
	BucY(Buckets),	\Y coordinate of top of bucket	   (0,0 = center)
	Ang(Buckets);	\angle of bucket around waterwheel (radians)

int	View,	\view selector (ScnA or ScnB)
	B,	\bucket number (0..Buckets-1)
	BX, BY,	\coordinates for a bucket (pixels)
	SX, SY,	\coordinates for plotting speed graph (pixels)
	SX0,	\steps per pixel for SX
	WLev;	\depth of water level in a bucket (pixels)

inc	c:\cxpl\codesi;		\include code definitions for intrinsic routines

def	Bel=$07, Esc=$1B, Sp=$20, Ctrl=$40;	\ASCII control characters

def	Black, Blue, Green, Cyan, Red, Magenta, Brown, White,	\attribute colors
	Gray, LBlue, LGreen, LCyan, LRed, LMagenta, Yellow, BWhite; \ EGA palette

def	Pi = 3.14159265358979323846, Pi2 = Pi*2.0;

\------------------------- STANDARD LIBRARY ROUTINES ---------------------------

func	Ror(N, C);		\Rotate N right C bits
int	N, C;
def	M = $000F;		\mask for 16-bit integers
return N>>(C&M) ! N<<(16-(C&M));



func	CallInt(Int, AX, BX, CX, DX, BP, DS, ES); \Call software interrupt
int	Int, AX, BX, CX, DX, BP, DS, ES; \(unused arguments need not be passed)
int	Cpureg;			\address of CPU register array
begin
Cpureg:= Getreg;
Cpureg(0):= AX;
Cpureg(1):= BX;
Cpureg(2):= CX;
Cpureg(3):= DX;
Cpureg(6):= BP;
Cpureg(9):= DS;
Cpureg(11):= ES;
Softint(Int);
return Cpureg(0);		\return AX register
end;	\CallInt



func	GetKey;			\Get character from keyboard (wait if necessary)
int	SC, Ch;			\this is a low-level routine with no echo,
begin				\ no Ctrl-C, and no cursor.
SC:= CallInt($16, $0000);	\function $00
Ch:= SC & $FF;
if Ch = 0 then Ch:= -(SC>>8);	\return non-ASCII chars as negative scan code
return Ch;
end;	\GetKey

\-------------------------------------------------------------------------------

proc	FlipView;		\Flip between video screens
begin
\Flip the Start Address Register between screens A and B.
\Note that the effect of this register occurs after the next vertical retrace.
if View = ScnB then
	begin
	View:= ScnA;
	port($3D4):= $0D;	\display screen A (ScnA)
	port($3D5):= $10;	\(140+1)*640/8 = $2C10
	port($3D4):= $0C;
	port($3D5):= $2C;
	end
else	begin
	View:= ScnB;
	port($3D4):= $0D;	\display screen B (ScnB)
	port($3D5):= $10;	\(140+1+339+13)*640/88 = $9A10
	port($3D4):= $0C;
	port($3D5):= $9A;
	end;
end;	\FlipView



proc	CTxt2(X, Y, Str);	\Display a text string at cursor position X, Y
int	X, Y;			\This displays on both video screens
char	Str;
begin
Cursor(X, Y+9);
Text(6, Str);
Cursor(X, Y+31);
Text(6, Str);
end;	\CTxt2



proc	CRlout2(X, Y, Num);	\Display a real number at cursor position X, Y
int	X, Y;			\This displays on both video screens
real	Num;
begin
Cursor(X, Y+9);
Rlout(6, Num);
Cursor(X, Y+31);
Rlout(6, Num);
end;	\CRlout2



proc	DrawLine(X0, Y0, X1, Y1, C);	\Draw a straight line
\Draw into the portion of video memory that is not currently being viewed
int	X0, Y0, X1, Y1, C;	\coordiantes, color
def	H=-5-16, V=40;		\centering adjustment (8 pixels per char cell)
begin
if View = ScnB then
	begin
	Move(X0+H, Y0+ScnA+V);
	Line(X1+H, Y1+ScnA+V, C);
	end
else	begin
	Move(X0+H, Y0+ScnB+V);
	Line(X1+H, Y1+ScnB+V, C);
	end
end;	\DrawLine



proc	DrawBlock(X0, Y0, X1, Y1, C);	\Draw a filled rectangle
int	X0, Y0, X1, Y1, C;	\coordinates of opposite corners, fill color
int	I, J, T;
begin
if X1 < X0 then [T:=X1;   X1:=X0;   X0:=T];	\put coordinates into
if Y1 < Y0 then [T:=Y1;   Y1:=Y0;   Y0:=T];	\ ascending order

if (Y1-Y0) > 8*(X1-X0) then		\if tall and skinny then
	for I:= X0, X1 do		\ draw vertical lines
		DrawLine(I, Y0, I, Y1, C)
else	for J:= Y0, Y1 do		\ else draw fast horizontal lines
		DrawLine(X0, J, X1, J, C);
end;	\DrawBlock



proc	DrawBucket(X0, Y0, X1, Y1, C);	\Draw a bucket
int	X0, Y0,		\coordinate of upper-left corner
	X1, Y1,		\coordiante of lower-right corner
	C;		\color
begin
DrawLine(X0, Y0, X0, Y1, C);
DrawLine(X0, Y1, X1, Y1, C);
DrawLine(X1, Y1, X1, Y0, C);
end;	\DrawBucket

\-------------------------------------------------------------------------------

begin	\Main
TrapC(true);	\(prevents Ctrl-C'ing out of a program in graphic-screen mode)
Format(3,1);	\display format of floating point numbers = XXX.X

\Set initial conditions
Speed:= 0.04;
WaterRate:= 20.0;
Fric:= 0.02;
Leak:= 0.006;
for B:= 0, Buckets-1 do Water(B):= 0.0;	\no water in buckets

Ang0:= 0.0;	\angle of reference point on wheel for bucket 0
AngB:= Pi2 / Float(Buckets);	\angle between buckets

SX:= 0;		\start speed graph at left edge of screen
SX0:= 0;

Setvid($12);	\set display for 640x480x16 graphics
View:= ScnA;

\Set up the Line Compare Register (= index $18 under the CRT Controller
\ Register = port $3D4) for split-screen operation. This causes the Horizontal
\ Scan Counter Register to be reset to zero when the video beam reaches scan
\ line 338 (the 339th line). Thus the next (340th) scan line displays the
\ contents of video memory starting at $A000:0. The low 8 bits of the Line
\ Compare Register are at index $18, but bit 8 is in index register 7 and bit 9
\ is in index register 9 (it's a real mess).

port($3D4):= $18;	\select Line Compare Register
port($3D5):= $52;	\set low 8 bits of 338 (=$152)

\Clear bit 9 of Line Compare Reg by clearing bit 6 of Maximum Scan Line Reg
port($3D4):= $9;
port($3D5):= port($3D5) & ~$40;

\(Bit 8 of Line Compare Reg is already set, so no code is needed)

Attrib(LCyan);
CTxt2(28, 0, "LORENZIAN WATERWHEEL");

Attrib(White);
CTxt2( 2, 2, "[S]peed:     ");
CTxt2( 2, 3, "[F]riction:       %");
CTxt2( 2, 4, "[W]ater Rate:");
CTxt2( 2, 5, "[L]eak Rate:      %");

CTxt2(54, 2, "Uppercase increases value");
CTxt2(54, 3, "Lowercase decreases value");
CTxt2(54, 4, "Spacebar empties buckets");
CTxt2(54, 5, "Esc terminates program");

Attrib(Cyan);
CTxt2( 2, 19, "Speed:");	\label for graph

Move(0,Base); Line(639,Base,Cyan);		\draw baseline of speed graph

loop	begin
	Move(SX, Base-SLimit); Line(SX, Base+SLimit, Cyan);	\vert graph line

	Attrib(White);
	CRlout2(15, 2, Speed * 180.0/Pi); 	\degrees/unit_time
	CRlout2(15, 3, Fric*100.0);		\percent of slowdown
	CRlout2(15, 4, WaterRate);		\arbitrary units
	CRlout2(15, 5, Leak*100.0);		\percent of water in bucket

	Ang0:= Ang0 + Speed;	\rotate wheel as a function of speed
	while Ang0 < 0.0 do Ang0:= Ang0 + Pi2; 	\limit angle to one revolution
	while Ang0 > Pi2 do Ang0:= Ang0 - Pi2;

	Torque:= 0.0;				\initialize torque accumulator
	for B:= 0, Buckets-1 do			\for all of the buckets...
		begin
		Ang(B):= Ang0 + AngB*Float(B);	\bucket angle around wheel
		BucX(B):= Sin(Ang(B)) * 100.0;	\bucket coordinates
		BucY(B):= Cos(Ang(B)) * 100.0;

		\Fill bucket under faucet
		if abs(BucX(B))<=Float(B2Width) & BucY(B)>0.0 then
			Water(B):= Water(B) + WaterRate;

		\Leak water out (rate is proportional to amount of water)
		Water(B):= Water(B) * (1.0-Leak);

		WLev:= Fix(Water(B)/Float(BWidth));	\limit amount of water
		if WLev > BHeight then Water(B):= Float(BHeight*BWidth);
		if WLev < 0 then Water(B):= 0.0; \(negative water is ugly!)

		\Sum torques (proportional to amount of water * X coordinate)
		Torque:= Torque + Water(B) * BucX(B);
		end;

	\Change in speed is proportional to torque
	Speed:= Speed + Torque/80E6;
	\(Experimentally determined fudge factor; it keeps values reasonable)

	\Reduce speed as a function of friction
	Speed:= Speed * (1.0-Fric);
	\(Linear friction (subtraction) doesn't cause chaos; this is more like
	\ air friction)

	\Erase old wheel image
	DrawBlock(320-100-B2Width, 0, 320+100+B2Width, 200+20+BHeight, Black);

	\Draw a picture of the results
	Attrib(White);
	CTxt2(37, 2, "o");		\faucet

	Attrib(Brown);
	CTxt2(37, 11, "o");		\axle

	\Draw water from faucet
	DrawBlock(320-1, 3, 320+1, 75, Ror($63,SX)<<8 +LBlue);

	\Draw buckets
	for B:= 0, Buckets-1 do
		begin
		BX:= Fix(BucX(B)) + 320;
		BY:= 100 - Fix(BucY(B)) + 20;
		WLev:= Fix(Water(B)/Float(BWidth));
		DrawBlock(BX-B2Width, BY+BHeight, BX+B2Width, BY+BHeight-WLev, LBlue);
		DrawBucket(BX-B2Width, BY, BX+B2Width, BY+BHeight, 
			if B=0 then Yellow else Brown);
		end;

	FlipView;				\display what was just drawn
	repeat until port($3DA) & $08;		\wait for vertical retrace


	\Draw graph of speed vs. time
	Move(SX, Base-SLimit); Line(SX, Base+SLimit, Black);	\erase old value
	Point(SX, Base, Cyan);			\restore baseline

	SY:= Fix(Speed*500.0);			\coordinate for speed
	if SY > SLimit then SY:= SLimit;	\ (amplitude is arbitrary)
	if SY < -SLimit then SY:= -SLimit;	\ limits prevent screen glitches
	Point(SX, Base-SY, LGreen);		\make clockwise speed positive

	SX0:= SX0 + 1;
	if SX0 = 4 then
		begin
		SX0:= 0;
		SX:= SX + 1;			\advance to next point
		end;

	if SX >= 640 then SX:= 0;		\wrap around screen


	\Handle keyboard commands
	if Chkkey then
		begin
		case Getkey of
		  ^S:	Speed:= Speed + 0.01;	\speed adjustment
		  ^s:	Speed:= Speed - 0.01;	\ (sometimes it needs a nudge)
		  ^F:	Fric:= Fric + 0.001;	\friction adjustment
		  ^f:	Fric:= Fric - 0.001;
		  ^W:	WaterRate:= WaterRate + 0.1;
		  ^w:	WaterRate:= WaterRate - 0.1;
		  ^L:	Leak:= Leak + 0.001;
		  ^l:	Leak:= Leak - 0.001;
		  Sp:	for B:= 0, Buckets-1 do Water(B):= 0.0;	\empty buckets
		  ^C-Ctrl, Esc:	quit		\terminate program
		other	Chout(0,Bel);
		end;
	end;	\loop
Setvid($03);					\restore normal text display
end;	\Main
